"""
An action is a class which performs some additinonal processing to a time_slice
before each iteration.

Please see the help for the prototype action for implementation details.
"""

from coffee.actions.actions import Prototype, BlowupCutoff, Info
from coffee.actions.hdf_output import SimOutput
